"""Main entry point for testing the library."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from pathlib import Path

##############################################################################
# Textual imports.
from textual import on
from textual.app import App, ComposeResult
from textual.containers import Center, Horizontal
from textual.widgets import Button, Footer, Label

##############################################################################
# Local imports.
from textual_fspicker import FileOpen, FileSave, Filters, SelectDirectory


##############################################################################
class TestApp(App[None]):
    """A simple test application."""

    CSS = """
    Screen#_default {
        align: center middle;
    }

    Screen#_default Horizontal {
        align: center middle;
        height: auto;
        margin-bottom: 1;
    }

    Screen#_default Horizontal Button {
        margin-left: 1;
        margin-right: 1;
    }
    """

    def compose(self) -> ComposeResult:
        """Compose the layout of the test application."""
        with Horizontal():
            yield Button("Open a file", id="open")
            yield Button("Save a file", id="save")
            yield Button("Select a directory", id="directory")
        with Center():
            yield Label("Press the button to pick something")
        yield Footer()

    def show_selected(self, to_show: Path | None) -> None:
        """Show the file that was selected by the user.

        Args:
            to_show: The file to show.
        """
        self.query_one(Label).update("Cancelled" if to_show is None else str(to_show))

    @on(Button.Pressed, "#open")
    def open_file(self) -> None:
        """Show the `FileOpen` dialog when the button is pushed."""
        self.push_screen(
            FileOpen(
                ".",
                filters=Filters(
                    ("Python", lambda p: p.suffix.lower() == ".py"),
                    ("Any", lambda _: True),
                    ("Emacs", lambda p: p.suffix.lower() == ".el"),
                    ("Lisp", lambda p: p.suffix.lower() in (".lisp", ".lsp", ".cl")),
                    ("Pascal", lambda p: p.suffix.lower() == ".pas"),
                    ("Clipper", lambda p: p.suffix.lower() in (".prg", ".ch")),
                    ("C", lambda p: p.suffix.lower() in (".c", ".h")),
                    ("C++", lambda p: p.suffix.lower() in (".cpp", ".cc", ".h")),
                ),
            ),
            callback=self.show_selected,
        )

    @on(Button.Pressed, "#save")
    def save_file(self) -> None:
        """Show the `FileSave` dialog when the button is pushed."""
        self.push_screen(FileSave(can_overwrite=False), callback=self.show_selected)

    @on(Button.Pressed, "#directory")
    def select_directory(self) -> None:
        """show the `SelectDirectory` dialog when the button is pushed."""
        self.push_screen(SelectDirectory(), callback=self.show_selected)


##############################################################################
if __name__ == "__main__":
    TestApp().run()

### __main__.py ends here



"""Provides a directory selection dialog."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from pathlib import Path

##############################################################################
# Textual imports.
from textual import on
from textual.app import ComposeResult
from textual.widgets import Button, Label

##############################################################################
# Local imports.
from .base_dialog import FileSystemPickerScreen
from .parts import DirectoryNavigation


##############################################################################
class SelectDirectory(FileSystemPickerScreen):
    """A directory selection dialog."""

    DEFAULT_CSS = """
    SelectDirectory > Dialog > InputBar > Label {
        width: 1fr;
        border: tall $background;
        padding-left: 1;
        padding-right: 1;
    }
    """

    def __init__(
        self, location: str | Path = ".", title: str = "Select directory"
    ) -> None:
        """Initialise the dialog.

        Args:
            location: Optional starting location.
            title: Optional title.
        """
        super().__init__(location, title)

    def on_mount(self) -> None:
        """Configure the dialog once the DOM is ready."""
        navigation = self.query_one(DirectoryNavigation)
        navigation.show_files = False
        self._set_current(navigation.location)

    def _input_bar(self) -> ComposeResult:
        """Provide any widgets for the input before, before the buttons."""
        yield Label()

    def _set_current(self, location: Path) -> None:
        """Set the current location.

        Args:
            location: The location to indicate.
        """
        current_selection = self.query_one("InputBar > Label", Label)
        # TODO: A nicer way of indicating we're looking at just the tail.
        current_selection.update(str(location)[-current_selection.size.width :])

    @on(DirectoryNavigation.Changed)
    def _show_selected(self, event: DirectoryNavigation.Changed) -> None:
        """Update the display of the current location.

        Args:
            event: The event with the selection information in.
        """
        event.stop()
        self._set_current(event.control.location)

    @on(Button.Pressed, "#select")
    def _select_directory(self, event: Button.Pressed) -> None:
        """React to the select button being pressed.

        Args:
            event: The button press event.
        """
        event.stop()
        self.dismiss(result=self.query_one(DirectoryNavigation).location)


### select_directory.py ends here



"""Defines the parts that make up the filesystem picker dialogs."""

##############################################################################
# Local imports.
from .directory_navigation import DirectoryNavigation
from .drive_navigation import DriveNavigation

##############################################################################
# Export public items.
__all__ = ["DirectoryNavigation", "DriveNavigation"]

### __init__.py ends here



"""Base file-oriented dialog."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
import sys
from pathlib import Path

##############################################################################
# Textual imports.
from textual import on
from textual.app import ComposeResult
from textual.events import Mount
from textual.widgets import Button, Input, Select

##############################################################################
# Local imports.
from .base_dialog import FileSystemPickerScreen
from .parts import DirectoryNavigation, DriveNavigation
from .path_filters import Filters
from .path_maker import MakePath


##############################################################################
class FileFilter(Select[int]):
    """The file filter."""


##############################################################################
class BaseFileDialog(FileSystemPickerScreen):
    """The base dialog for file-oriented picking dialogs."""

    DEFAULT_CSS = """
    BaseFileDialog InputBar Input {
        width: 2fr;
    }

    BaseFileDialog InputBar Select {
        width: 1fr;
    }
    """

    def __init__(
        self,
        location: str | Path = ".",
        title: str = "Open",
        select_button: str = "",
        *,
        filters: Filters | None = None,
    ) -> None:
        """Initialise the base dialog dialog.

        Args:
            location: Optional starting location.
            title: Optional title.
            select_button: The label for the selection button.
            filters: Optional filters to show in the dialog.
        """
        super().__init__(location, title, select_button=select_button)
        self._filters = filters
        """The filters for the dialog."""

    def _input_bar(self) -> ComposeResult:
        """Provide any widgets for the input before, before the buttons."""
        yield Input()
        if self._filters:
            yield FileFilter(
                self._filters.selections,
                prompt="File filter",
                value=0,
                allow_blank=False,
            )

    @on(Mount)
    def _initial_filter(self) -> None:
        """Set the initial filter once the DOM is ready."""
        if self._filters:
            self.query_one(DirectoryNavigation).file_filter = self._filters[0]

    @on(DirectoryNavigation.Selected)
    def _select_file(self, event: DirectoryNavigation.Selected) -> None:
        """Handle a file being selected in the picker.

        Args:
            event: The event to handle.
        """
        file_name = self.query_one(Input)
        file_name.value = str(event.path.name)
        file_name.focus()

    @on(Input.Changed)
    def _clear_error(self) -> None:
        """Clear any error that might be showing."""
        super()._clear_error()

    @on(Select.Changed)
    def _change_filter(self, event: Select.Changed) -> None:
        """Handle a change in the filter.

        Args:
            event: The event to handle.
        """
        if self._filters is not None and isinstance(event.value, int):
            self.query_one(DirectoryNavigation).file_filter = self._filters[event.value]
        else:
            self.query_one(DirectoryNavigation).file_filter = None
        self.query_one(DirectoryNavigation).focus()

    def _should_return(self, candidate: Path) -> bool:
        """Final check on a picked file before returning it to the caller.

        Args:
            candidate: The file to check.

        Returns:
            `True` if the file should be returned, `False` if not.

        Note:
            This method is designed to be called as a final check; this is a
            good place to set up the display of an error before returning
            `False`, for example.
        """
        del candidate
        return True

    @on(Input.Submitted)
    @on(Button.Pressed, "#select")
    def _confirm_file(self, event: Input.Submitted | Button.Pressed) -> None:
        """Confirm the selection of the file in the input box.

        Args:
            event: The event to handle.
        """
        event.stop()
        file_name = self.query_one(Input)

        # Only even try and process this if there's some input.
        if not file_name.value:
            self._set_error("A file must be chosen")
            return

        # If it looks like the user is typing in some sort of home
        # directory path... (does pathlib let me test for this, or at
        # least ask what the home character is? Docs don't mention this;
        # so for now I'm going to hard-code this).
        if file_name.value.startswith("~"):
            # ...let's simply expand and go with that.
            try:
                chosen = MakePath.of(file_name.value).expanduser()
            except RuntimeError as error:
                self._set_error(str(error))
                return
        else:
            # It's not a home directory path, so let's combine with the
            # location of the directory navigator widget.
            chosen = (
                self.query_one(DirectoryNavigation).location / file_name.value
            ).resolve()

        # If it's a directory, approach it like it's the user simply
        # doing a "cd".
        try:
            if chosen.is_dir():
                if sys.platform == "win32":
                    if drive := MakePath.of(file_name.value).drive:
                        self.query_one(DriveNavigation).drive = drive
                self.query_one(DirectoryNavigation).location = chosen
                self.query_one(DirectoryNavigation).focus()
                self.query_one(Input).value = ""
                return
        except PermissionError:
            self._set_error("Permission error")
            return

        # If the chosen file passes the final tests...
        if self._should_return(chosen):
            # ...return it.
            self.dismiss(result=chosen)


### file_dialog.py ends here



"""A library that provides widgets for selecting things from the filesystem."""

##############################################################################
# Python imports.
from importlib.metadata import version

######################################################################
# Main app information.
__author__ = "Dave Pearson"
__copyright__ = "Copyright 2023, Dave Pearson"
__credits__ = ["Dave Pearson"]
__maintainer__ = "Dave Pearson"
__email__ = "davep@davep.org"
__version__ = version("textual_fspicker")
__licence__ = "MIT"

##############################################################################
# Local imports.
from .file_open import FileOpen
from .file_save import FileSave
from .path_filters import Filters
from .path_maker import MakePath
from .select_directory import SelectDirectory

##############################################################################
# Export the imports.
__all__ = ["FileOpen", "FileSave", "SelectDirectory", "Filters", "MakePath"]

### __init__.py ends here



"""Support functions for doing safe filesystem tests.

This module provides some functions for performing tests on entries in the
filesystem in a way that swallows some errors that we don't want causing the
application to crash on. In most cases this is going to be a
PermissionError, which for UI stuff is better ignored than raised.
"""

##############################################################################
# Python imports.
from pathlib import Path


##############################################################################
def is_dir(location: Path) -> bool:
    """A safe version of is_dir.

    Args:
        location: The location to test.

    Returns:
        `True` if the location looks like a directory, `False` if not, or if
        it could not be determined.

    Note:
        This function swallows `PermissionError` and just returns that the
        location isn't a directory.
    """
    try:
        return location.is_dir()
    except PermissionError:
        return False


##############################################################################
def is_file(location: Path) -> bool:
    """A safe version of is_file.

    Args:
        location: The location to test.

    Returns:
        `True` if the location looks like a file or if it could not be
        determined, `False` it can be known it doesn't look like a file.

    Note:
        This function swallows `PermissionError` and just returns that the
        location is a file.
    """
    try:
        return location.is_file()
    except PermissionError:
        return True


##############################################################################
def is_symlink(location: Path) -> bool:
    """A safe version of is_symlink.

    Args:
        location: The location to test.

    Returns:
        `True` if the location looks like a symlink, `False` if not, or if
        it could not be determined.

    Note:
        This function swallows `PermissionError` and just returns that the
        location isn't a symlink.
    """
    try:
        return location.is_symlink()
    except PermissionError:
        return False


### safe_tests.py ends here



"""Code to handle the filters for the dialogs."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from pathlib import Path
from typing import Callable, NamedTuple

##############################################################################
# Typing extension imports.
from typing_extensions import TypeAlias

##############################################################################
FilterFunction: TypeAlias = Callable[[Path], bool]
"""Type of a path filter function."""


##############################################################################
class Filter(NamedTuple):
    """A path filter."""

    name: str
    """The name of the filter."""

    tester: FilterFunction
    """The tester for the filter."""

    def __call__(self, path: Path) -> bool:
        return self.tester(path)


##############################################################################
class Filters:
    """A path filter collection."""

    def __init__(self, *filters: Filter | tuple[str, FilterFunction]) -> None:
        """Initialise the filter collection.

        Args:
            filters: The initial set of filters.
        """
        self._filters: list[Filter] = list(
            path_filter if isinstance(path_filter, Filter) else Filter(*path_filter)
            for path_filter in filters
        )

    @property
    def selections(self) -> list[tuple[str, int]]:
        """Get the filters in a `Select`-friendly way."""
        return [
            (file_filter.name, filter_id)
            for filter_id, file_filter in enumerate(self._filters)
        ]

    def __getitem__(self, filter_id: int) -> Filter:
        return self._filters[filter_id]

    def __bool__(self) -> bool:
        return bool(self._filters)


### path_filters.py ends here



"""Provides a widget for drive navigation."""

##############################################################################
# Python imports.
import sys
from dataclasses import dataclass
from pathlib import Path

##############################################################################
# Import or provide a version of listdrives.
try:
    from os import listdrives  # type: ignore[attr-defined]
except ImportError:
    import string

    def listdrives() -> list[str]:
        """Return a list containing the names of drives in the system.

        A drive name typically looks like 'C:\\'. This is an implementation for
        Python versions before 3.12. It does _not_ list removable drives which
        have no media inserted.

        Returns:
            list[str]: The list of available drives.
        """
        return [
            f"{letter}:"
            for letter in string.ascii_uppercase
            if Path(f"{letter}:\\").exists()
        ]


##############################################################################
# Textual imports.
from textual import on
from textual.message import Message
from textual.reactive import var
from textual.widgets import OptionList
from textual.widgets.option_list import Option

##############################################################################
# Local imports.
from ..path_maker import MakePath


##############################################################################
class DriveEntry(Option):
    """A drive entry for the `DriveNavigation` class."""

    def __init__(self, drive: Path | str) -> None:
        """Initialise the object.

        Args:
            drive: The drive to handle.
        """
        self.drive_root: Path = MakePath.of(drive)
        """The drive root for this entry."""
        super().__init__(self.drive_root.drive, id=self.drive_root.drive)


##############################################################################
class DriveNavigation(OptionList):
    """A drive navigation widget.

    Provides a single-pane widget that lets the user select a drive. This is
    very useful in combination with the `DirectoryNavigation` widget. A dialog can
    reload that widget in response to drive selection changes.
    """

    DEFAULT_CSS = """
    DriveNavigation, DriveNavigation:focus {
        border: blank;
        border-right: $panel-darken-1;
        width: 10;
        height: 100%;
    }

    DriveNavigation > .option-list--option-highlighted {
        text-style: bold;
        color: white;
    }
    """
    """Default styling for the widget."""

    @dataclass
    class DriveSelected(Message):
        """Message sent when a drive is selected."""

        drive_root: Path
        """The selected drive root, like `c:\\`."""

    drive: var[str] = var[str](MakePath.of(".").absolute().drive, init=False)

    def __init__(self, location: Path | str = ".") -> None:
        """Initialise the drive navigation widget.

        Args:
            location: The starting location.
        """
        super().__init__()
        self.set_reactive(DriveNavigation.drive, MakePath.of(location).absolute().drive)
        if sys.platform == "win32":
            self._entries = [DriveEntry(drive) for drive in listdrives()]
        else:
            self._entries: list[DriveEntry] = []

    def on_mount(self) -> None:
        """Add available drives to the widget."""
        self.add_options(self._entries)
        self.highlight_drive(self.drive)

    def _watch_drive(self, drive: str) -> None:
        """Highlight the new drive.

        Args:
            drive: The new value of the current drive.
        """
        self.highlight_drive(drive)

    def highlight_drive(self, drive: str) -> None:
        """Highlight the given drive.

        Args:
            drive: The drive to be highlighted.
        """
        self.highlighted = self.get_option_index(drive.upper())

    @on(OptionList.OptionSelected)
    def drive_selected(self, event: OptionList.OptionSelected) -> None:
        """Post a DriveSelected message.

        Args:
            event: The drive selected event from the parent `OptionList`.
        """
        assert isinstance(event.option, DriveEntry)
        event.stop()
        self.drive = event.option.drive_root.drive
        self.post_message(self.DriveSelected(drive_root=event.option.drive_root))


### drive_navigation.py ends here



"""Provides a file save dialog."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from pathlib import Path

##############################################################################
# Local imports.
from .file_dialog import BaseFileDialog
from .path_filters import Filters


##############################################################################
class FileSave(BaseFileDialog):
    """A file save dialog."""

    def __init__(
        self,
        location: str | Path = ".",
        title: str = "Save as",
        *,
        filters: Filters | None = None,
        can_overwrite: bool = True,
    ) -> None:
        """Initialise the `FileOpen` dialog.

        Args:
            location: Optional starting location.
            title: Optional title.
            filters: Optional filters to show in the dialog.
            can_overwrite: Flag to say if an existing file can be overwritten.
        """
        super().__init__(location, title, select_button="Save", filters=filters)
        self._can_overwrite = can_overwrite
        """Can an existing file be overwritten?"""

    def _should_return(self, candidate: Path) -> bool:
        """Perform the final checks on the chosen file.

        Args:
            candidate: The file to check.
        """
        if candidate.exists() and not self._can_overwrite:
            self._set_error("Overwrite is not allowed")
            return False
        return True


### file_save.py ends here



"""The base dialog code for the other dialogs in the library."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
import sys
from pathlib import Path
from typing import Optional

##############################################################################
# Textual imports.
from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button

##############################################################################
# Local imports.
from .parts import DirectoryNavigation, DriveNavigation


##############################################################################
class Dialog(Vertical):
    """Layout class for the main dialog area."""


##############################################################################
class InputBar(Horizontal):
    """The input bar area of the dialog."""


##############################################################################
class FileSystemPickerScreen(ModalScreen[Optional[Path]]):
    """Base screen for the dialogs in this library."""

    DEFAULT_CSS = """
    FileSystemPickerScreen {
        align: center middle;
    }

    FileSystemPickerScreen Dialog {
        width: 80%;
        height: 80%;
        border: $border;
        background: $panel;
        border-title-color: $text;
        border-title-background: $panel;
        border-subtitle-color: $text;
        border-subtitle-background: $error;

        OptionList, OptionList:focus {
            background: $panel;
            background-tint: $panel;
        }
    }

    FileSystemPickerScreen DirectoryNavigation {
        height: 1fr;
    }

    FileSystemPickerScreen InputBar {
        height: auto;
        align: right middle;
        padding-top: 1;
        padding-right: 1;
        padding-bottom: 1;
    }

    FileSystemPickerScreen InputBar Button {
        margin-left: 1;
    }
    """

    BINDINGS = [Binding("full_stop", "hidden"), Binding("escape", "dismiss(None)")]
    """The bindings for the dialog."""

    def __init__(
        self, location: str | Path = ".", title: str = "", select_button: str = ""
    ) -> None:
        """Initialise the dialog.

        Args:
            location: Optional starting location.
            title: Optional title.
            select_button: Label for the select button.
        """
        super().__init__()
        self._location = location
        """The starting location."""
        self._title = title
        """The title for the dialog."""
        self._select_button = select_button or "Select"
        """The text prompt for the select button."""

    def _input_bar(self) -> ComposeResult:
        """Provide any widgets for the input bar, before the buttons."""
        yield from ()

    def compose(self) -> ComposeResult:
        """Compose the child widgets.

        Returns:
            The widgets to compose.
        """
        with Dialog() as dialog:
            dialog.border_title = self._title
            with Horizontal():
                if sys.platform == "win32":
                    yield DriveNavigation(self._location)
                yield DirectoryNavigation(self._location)
            with InputBar():
                yield from self._input_bar()
                yield Button(self._select_button, id="select")
                yield Button("Cancel", id="cancel")

    def on_mount(self) -> None:
        """Focus directory widget on mount."""
        self.query_one(DirectoryNavigation).focus()

    def _set_error(self, message: str = "") -> None:
        """Set or clear the error message.

        Args:
            message: Optional message to show as an error.
        """
        self.query_one(Dialog).border_subtitle = message

    @on(DriveNavigation.DriveSelected)
    def _change_drive(self, event: DriveNavigation.DriveSelected) -> None:
        """Reload DirectoryNavigation in response to drive change."""
        self.query_one(DirectoryNavigation).location = event.drive_root

    @on(DirectoryNavigation.Changed)
    def _clear_error(self) -> None:
        """Clear any error that might be showing."""
        self._set_error()

    @on(DirectoryNavigation.PermissionError)
    def _show_permission_error(self) -> None:
        """Show any permission error bubbled up from the directory navigator."""
        self._set_error("Permission error")

    @on(Button.Pressed, "#cancel")
    def _cancel(self, event: Button.Pressed) -> None:
        """Cancel the dialog.

        Args:
            event: The even to handle.
        """
        event.stop()
        self.dismiss(None)

    def action_hidden(self) -> None:
        """Action for toggling the display of hidden entries."""
        self.query_one(DirectoryNavigation).toggle_hidden()


### base_dialog.py ends here
